/*
 * Copyright (c) 2012, Masanori Miyamoto.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1.Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2.Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE FREEBSD PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */ 
#include <stdio.h>
#include <stdlib.h>
#include "CzarBasic_vm.h"
#include "CzarBasic.h"
#include "CzarBasic_error.h"
#include "lex_analyzer.h"
#include "CzFStream.h"
#include "CzMemStream.h"
#include "parser.h"

#define VAR_SIZE	26	/* 変数の数 */
#define STACK_SIZE	256	/* スタックサイズ */

/* プログラムコードメモリ */
int g_vmmem[VMMEM_MAX];
int g_vmmem_last;
/* 変数 */
static int st_var[VAR_SIZE];
/* スタック */
static int st_stack[STACK_SIZE];
/* スタックポインタ */
static int st_sp;
/* プログラムカウンタ */
int g_pc;
/* 文字列プール */
char g_string_pool[STRING_POOL_MAX];
/* 行番号テーブル */
struct line_num_table g_line_num_table[LINE_NUM_MAX];

/* 現在のエディタでの行番号 */
static int st_edit_line;
/* 現在のベーシックでの行番号 */
static int st_basic_line;

static void push_val(int val);
static int pop_val();
static void inc_pc();
static int search_line_num(int line_num);

static void op_push_int();
static void op_push_var();
static void op_pop_var();
static void op_add();
static void op_sub();
static void op_mul();
static void op_div();
static void op_cmp_eq();
static void op_cmp_ne();
static void op_cmp_gt();
static void op_cmp_ge();
static void op_cmp_lt();
static void op_cmp_le();
static void op_cond_jmp();
static void op_jmp();
static void op_jmp_num();
static void op_call();
static void op_ret();
static void op_input();
static void op_string();
static void op_prn();
static void op_clear();
static void op_end();
static void op_ncond_jmp();
static void op_ld_el();
static void op_ld_bl();

/* vm実行 */
void vm_run()
{
	g_pc = 0;
	st_sp = 0;

	while(1) {
		int code = g_vmmem[g_pc];
		inc_pc();
		switch((enum opecode)code) {
			case OP_PUSH_INT:	/* 数値をスタックに積む */
				op_push_int();
				break;
			case OP_PUSH_VAR:	/* 変数をスタックに積む */
				op_push_var();
				break;
			case OP_POP_VAR:	/* 変数にスタックから値を代入 */
				op_pop_var();
				break;
			case OP_ADD:		/* 加算 */
				op_add();
				break;
			case OP_SUB:		/* 減算 */
				op_sub();
				break;
			case OP_MUL:		/* 乗算 */
				op_mul();
				break;
			case OP_DIV:		/* 除算 */
				op_div();
				break;
			case OP_CMP_EQ:		/* 比較(=) */
				op_cmp_eq();
				break;
			case OP_CMP_NE:		/* 比較(<>) */
				op_cmp_ne();
				break;
			case OP_CMP_GT:		/* 比較(>) */
				op_cmp_gt();
				break;
			case OP_CMP_GE:		/* 比較(>=) */
				op_cmp_ge();
				break;
			case OP_CMP_LT:		/* 比較(<) */
				op_cmp_lt();
				break;
			case OP_CMP_LE:		/* 比較(<=) */
				op_cmp_le();
				break;
			case OP_COND_JMP:	/* 条件ジャンプ */
				op_cond_jmp();
				break;
			case OP_JMP:		/* 無条件ジャンプ */
				op_jmp();
				break;
			case OP_JMP_NUM:	/* 行番号ジャンプ */
				op_jmp_num();
				break;
			case OP_CALL:		/* 行番号サブルーチンコール */
				op_call();
				break;
			case OP_RET:		/* サブルーチンコールリターン */
				op_ret();
				break;
			case OP_INPUT:		/* INPUTステートメント */
				op_input();
				break;
			case OP_PRN_STRING:	/* 文字列出力 */
				op_string();
				break;
			case OP_PRN:		/* スタック上の値を出力 */
				op_prn();
				break;
			case OP_CLEAR:		/* 変数のクリア */
				op_clear();
				break;
			case OP_END:		/* 終了 */
				op_end();
				break;
			case OP_NCOND_JMP:	/* 偽のときジャンプ */
				op_ncond_jmp();
				break;
			case OP_LD_EL:		/* エディタでの行番号 */
				op_ld_el();
				break;
			case OP_LD_BL:		/* BASICでの行番号 */
				op_ld_bl();
				break;
			default:
				break;
		}
	}
}

static void push_val(int val)
{
	st_stack[st_sp] = val;
	st_sp++;
	if(st_sp >= STACK_SIZE) {
		ErrorExit(ERR_STACK_OVERFLOW, st_edit_line, st_basic_line);
	}
}

static int pop_val()
{
	st_sp--;
	if(st_sp < 0) {
		ErrorExit(ERR_STACK_UNDERFLOW, st_edit_line, st_basic_line);
	}
	int ret = st_stack[st_sp];
	return ret;
}

static void inc_pc()
{
	g_pc++;
	if(g_pc >= VMMEM_MAX) {
		g_pc = 0;
	}
}

static int search_line_num(int line_num)
{
	int i = 0;
	while(g_line_num_table[i].line_num != -1) {
		if(g_line_num_table[i].line_num == line_num) {
			return g_line_num_table[i].address;
		}
		i++;
	}
	return -1;
}

static void op_push_int()
{
	push_val(g_vmmem[g_pc]);
	inc_pc();
}

static void op_push_var()
{
	if(g_vmmem[g_pc] < 0 || g_vmmem[g_pc] >= VAR_SIZE) {
		ErrorExit(ERR_OUT_OF_RANGE, st_edit_line, st_basic_line);
	}
	push_val(st_var[g_vmmem[g_pc]]);
	inc_pc();
}

static void op_pop_var()
{
	if(g_vmmem[g_pc] < 0 || g_vmmem[g_pc] >= VAR_SIZE) {
		ErrorExit(ERR_OUT_OF_RANGE, st_edit_line, st_basic_line);
	}
	st_var[g_vmmem[g_pc]] = pop_val();
	inc_pc();
}

static void op_add()
{
	int v2 = pop_val();
	int v1 = pop_val();
	int val = v1 + v2;
	if(val > 32767) {
		ErrorExit(ERR_OVERFLOW, st_edit_line, st_basic_line);
	}
	push_val(val);
}

static void op_sub()
{
	int v2 = pop_val();
	int v1 = pop_val();
	int val = v1 - v2;
	if(val < -32768) {
		ErrorExit(ERR_UNDERFLOW, st_edit_line, st_basic_line);
	}
	push_val(val);
}

static void op_mul()
{
	int v2 = pop_val();
	int v1 = pop_val();
	int val = v1 * v2;
	if(val > 32767) {
		ErrorExit(ERR_OVERFLOW, st_edit_line, st_basic_line);
	}
	push_val(val);
}

static void op_div()
{
	int v2 = pop_val();
	int v1 = pop_val();
	if(v2 == 0) {
		ErrorExit(ERR_DIVISIONBYZERO, st_edit_line, st_basic_line);
	}
	push_val(v1 / v2);
}

static void op_cmp_eq()
{
	int v2 = pop_val();
	int v1 = pop_val();
	push_val(v1 == v2 ? -1:0);
}

static void op_cmp_ne()
{
	int v2 = pop_val();
	int v1 = pop_val();
	push_val(v1 != v2 ? -1:0);
}

static void op_cmp_gt()
{
	int v2 = pop_val();
	int v1 = pop_val();
	push_val(v1 > v2 ? -1:0);
}

static void op_cmp_ge()
{
	int v2 = pop_val();
	int v1 = pop_val();
	push_val(v1 >= v2 ? -1:0);
}

static void op_cmp_lt()
{
	int v2 = pop_val();
	int v1 = pop_val();
	push_val(v1 < v2 ? -1:0);
}

static void op_cmp_le()
{
	int v2 = pop_val();
	int v1 = pop_val();
	push_val(v1 <= v2 ? -1:0);
}

static void op_cond_jmp()
{
	int val = pop_val();

	if(val == -1) {
		g_pc = g_vmmem[g_pc];
		return;
	}
	inc_pc();
}

static void op_jmp()
{
	g_pc = g_vmmem[g_pc];
}

static void op_jmp_num()
{
	int line_num = pop_val();
	int addr = search_line_num(line_num);
	if(addr == -1) {
		ErrorExit(ERR_UNDEFINED_LINE_NUMBER, st_edit_line, st_basic_line);
	}
	g_pc = addr;
}

static void op_call()
{
	int line_num = pop_val();
	int addr = search_line_num(line_num);
	if(addr == -1) {
		ErrorExit(ERR_UNDEFINED_LINE_NUMBER, st_edit_line, st_basic_line);
	}
	push_val(g_pc);
	g_pc = addr;
}

static void op_ret()
{
	int ret = pop_val();
	g_pc = ret;
}

static void op_input()
{
	int i = 0;
	char buf[256];
	CZFSTREAM stream;
	CZMEMSTREAM memstream;
	TOKEN *token;

	int var_size = g_vmmem[g_pc];
	inc_pc();

	CzFStream_init(&stream, stdin);
	while(1) {
		fprintf(stdout, "? ");
		fflush(stdout);
		int line_size;
		/* 一行入力 */
		if((line_size = stream.super.lineinput((CZSTREAM*)&stream, buf, 256)) < 0) {
			ErrorExit(ERR_BUFFER_OVERFLOW, st_edit_line, st_basic_line);
		}
		CzMemStream_init(&memstream, buf, line_size);
		/* 1つ目のトークンから整数値を取得 */
		PARSE_RESULT ret;
		int retval;
		while(i != var_size) {
			token = lex_get_token((CZSTREAM*)&memstream);
			if(token->token_type == TOKEN_TYPE_EOF
					|| token->token_type == TOKEN_TYPE_NL) {
				break;
			}
			lex_unget_token(token);
			ret = parse_int((CZSTREAM*)&memstream, &retval);
			if(ret == PARSE_RESULT_FALSE) { /* 整数でないなら */
				i = 0;
				fprintf(stdout, "Redo from start\n");
				break;
			} else if(ret == PARSE_RESULT_TRUE) { /* 整数なら */
				st_var[st_stack[st_sp - var_size + i]] = retval;
				i++;
			} else {
				memstream.super.close((CZSTREAM*)&memstream);
				ErrorExit(ERR_BUFFER_OVERFLOW, st_edit_line, st_basic_line);
			}
			token = lex_get_token((CZSTREAM*)&memstream);
			if(token->token[0] != ',' && token->token[1] != 0) {
				lex_unget_token(token);
			}
		}
		if(i == var_size) {
			if(!memstream.super.eof((CZSTREAM*)&memstream)) { /* まだ入力項目があるなら */
				fprintf(stderr, "Extra ignored\n");
			}
			memstream.super.close((CZSTREAM*)&memstream);
			break;
		}
		memstream.super.close((CZSTREAM*)&memstream);
	}
	st_sp -= var_size;
}

static void op_string()
{
	int v = g_vmmem[g_pc];
	inc_pc();
	printf("%s", &g_string_pool[v]);
}

static void op_prn()
{
	int v = pop_val();
	printf("%d", v);
}

static void op_clear()
{
	int i;
	for(i = 0; i < VAR_SIZE; i++) {
		st_var[i] = 0;
	}
}

static void op_end()
{
	exit(0);
}

static void op_ncond_jmp()
{
	int val = pop_val();

	if(val == 0) {
		g_pc = g_vmmem[g_pc];
		return;
	}
	inc_pc();
}

/* エディタでの行番号を設定 */
static void op_ld_el()
{
	st_edit_line = g_vmmem[g_pc];
	inc_pc();
}

/* BASICでの行番号を設定 */
static void op_ld_bl()
{
	st_basic_line = g_vmmem[g_pc];
	inc_pc();
}
