/*
 * Copyright (c) 2012, Masanori Miyamoto.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1.Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2.Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE FREEBSD PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */ 
#include <stdio.h>
#include "CzarBasic.h"
#include "CzarBasic_vm.h"

typedef void (*PRINT_FUNC)();

static void parse_nop();
static void parse_push_int();		/* 数値をスタックに積む */
static void parse_push_var();	/* 変数をスタックに積む */
static void parse_pop_var();		/* 変数にスタックから値を代入 */
static void parse_add();			/* 加算 */
static void parse_sub();			/* 減算 */
static void parse_mul();			/* 乗算 */
static void parse_div();			/* 除算 */
static void parse_cmp_eq();		/* 比較(=) */
static void parse_cmp_ne();		/* 比較(<>) */
static void parse_cmp_gt();		/* 比較(>) */
static void parse_cmp_ge();		/* 比較(>=) */
static void parse_cmp_lt();	/* 比較(<) */
static void parse_cmp_le();		/* 比較(<=) */
static void parse_cond_jmp();		/* 条件ジャンプ */
static void parse_jmp();			/* 無条件ジャンプ */
static void parse_jmp_num();		/* 行番号ジャンプ */
static void parse_call();		/* 行番号サブルーチンコール */
static void parse_ret();			/* サブルーチンコールリターン */
static void parse_input();	/* INPUTステートメント */
static void parse_prn_string();		/* 文字列出力 */
static void parse_prn();			/* スタック上の値を出力 */
static void parse_clear();		/* 変数のクリア */
static void parse_end();			/* 終了 */
static void parse_ncond_jmp();		/* 条件(偽)ジャンプ */
static void parse_ld_el();
static void parse_ld_bl();

static PRINT_FUNC st_opecode_list[] = {
	parse_nop, 
	parse_push_int,		/* 数値をスタックに積む */
	parse_push_var,	/* 変数をスタックに積む */
	parse_pop_var,		/* 変数にスタックから値を代入 */
	parse_add,			/* 加算 */
	parse_sub,			/* 減算 */
	parse_mul,			/* 乗算 */
	parse_div,			/* 除算 */
	parse_cmp_eq,		/* 比較(=) */
	parse_cmp_ne,		/* 比較(<>) */
	parse_cmp_gt,		/* 比較(>) */
	parse_cmp_ge,		/* 比較(>=) */
	parse_cmp_lt,		/* 比較(<) */
	parse_cmp_le,		/* 比較(<=) */
	parse_cond_jmp,		/* 条件ジャンプ */
	parse_jmp,			/* 無条件ジャンプ */
	parse_jmp_num,		/* 行番号ジャンプ */
	parse_call,		/* 行番号サブルーチンコール */
	parse_ret,			/* サブルーチンコールリターン */
	parse_input,		/* INPUTステートメント */
	parse_prn_string,		/* 文字列出力 */
	parse_prn,			/* スタック上の値を出力 */
	parse_clear,		/* 変数のクリア */
	parse_end,			/* 終了 */
	parse_ncond_jmp,		/* 条件(偽)ジャンプ */
	parse_ld_el,
	parse_ld_bl,
};

/* ディスアセンブル */
void vm_disasm()
{
	for(g_pc = 0; g_pc < g_vmmem_last; g_pc++) {
		printf("%04X ", g_pc);
		st_opecode_list[g_vmmem[g_pc]]();
	}
}

static void parse_nop()
{
	printf("NOP\n");
}

/* 数値をスタックに積む */
static void parse_push_int()
{
	g_pc++;
	printf("PUSH_INT 0x%x\n", g_vmmem[g_pc]);
}
/* 変数をスタックに積む */
static void parse_push_var()
{
	g_pc++;
	printf("PUSH_VAR 0x%x\n", g_vmmem[g_pc]);
}
/* 変数にスタックから値を代入 */
static void parse_pop_var()
{
	g_pc++;
	printf("POP_VAR 0x%x\n", g_vmmem[g_pc]);
}
/* 加算 */
static void parse_add()
{
	printf("ADD\n");
}
/* 減算 */
static void parse_sub()
{
	printf("SUB\n");
}
/* 乗算 */
static void parse_mul()
{
	printf("MUL\n");
}
/* 除算 */
static void parse_div()
{
	printf("DIV\n");
}
/* 比較(=) */
static void parse_cmp_eq()
{
	printf("CMP_EQ\n");
}
/* 比較(<>) */
static void parse_cmp_ne()
{
	printf("CMP_NE\n");
}
/* 比較(>) */
static void parse_cmp_gt()
{
	printf("CMP_GT\n");
}
/* 比較(>=) */
static void parse_cmp_ge()
{
	printf("CMP_GE\n");
}
/* 比較(<) */
static void parse_cmp_lt()
{
	printf("CMP_LT\n");
}
/* 比較(<=) */
static void parse_cmp_le()
{
	printf("CMP_LE\n");
}
/* 条件ジャンプ */
static void parse_cond_jmp()
{
	g_pc++;
	printf("COND_JMP 0x%x\n", g_vmmem[g_pc]);
}
/* 無条件ジャンプ */
static void parse_jmp()
{
	g_pc++;
	printf("JMP 0x%x\n", g_vmmem[g_pc]);
}
/* 行番号ジャンプ */
static void parse_jmp_num()
{
	printf("JMP_NUM\n");
}

/* 行番号サブルーチンコール */
static void parse_call()
{
	printf("CALL\n");
}
/* サブルーチンコールリターン */
static void parse_ret()
{
	printf("RET\n");
}
/* INPUTステートメント */
static void parse_input()
{
	g_pc++;
	printf("INPUT %d\n", g_vmmem[g_pc]);
}
/* 文字列出力 */
static void parse_prn_string()
{
	g_pc++;
	printf("PRN_STRING %d\n", g_vmmem[g_pc]);
}
/* スタック上の値を出力 */
static void parse_prn()
{
	printf("PRN\n");
}
/* 変数のクリア */
static void parse_clear()
{
	printf("CLEAR\n");
}
/* 終了 */
static void parse_end()
{
	printf("END\n");
}

/* 条件(偽)ジャンプ */
static void parse_ncond_jmp()
{
	g_pc++;
	printf("NCOND_JMP 0x%x\n", g_vmmem[g_pc]);
}

static void parse_ld_el()
{
	g_pc++;
	printf("LD EL, %d\n", g_vmmem[g_pc]);
}

static void parse_ld_bl()
{
	g_pc++;
	printf("LD BL, %d\n", g_vmmem[g_pc]);
}
