/*
 * Copyright (c) 2012, Masanori Miyamoto.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1.Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2.Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE FREEBSD PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "parser.h"
#include "lex_analyzer.h"
#include "CzarBasic_vm.h"
#include "CzarBasic_error.h"

/* 文字列を比較 */
static int Strcmp(const char *str1, const char *str2);

/* tokenを符号無し整数に変換 */
static PARSE_RESULT conv_unsigned_int(TOKEN *token, int *retval);

/* 文字列プールポインタ */
static int st_string_pt;
/* 行番号テーブルポインタ */
static int st_table_pt;
/* 現在のテキスト行番号 */
static int st_now_line_num;
/* 現在のBASICでの行番号 */
static int st_now_basic_line_num;

/* 行解析 */
static void parse_line(CZSTREAM *stream);
/* 命令解析 */
static void parse_statement(CZSTREAM *stream);
/* 式解析 */
static void parse_expression(CZSTREAM *stream);

static void parse_print(CZSTREAM *stream);
static void parse_if(CZSTREAM *stream);
static void parse_goto(CZSTREAM *stream);
static void parse_let(CZSTREAM *stream);
static void parse_input(CZSTREAM *stream);
static void parse_gosub(CZSTREAM *stream);
static void parse_return(CZSTREAM *stream);
static void parse_clear(CZSTREAM *stream);
static void parse_end(CZSTREAM *stream);
/* 項1解析 */
static void parse_item1(CZSTREAM *stream);
/* 項2解析 */
static void parse_item2(CZSTREAM *stream);
/* 基本式解析 */
static void parse_basis_expression(CZSTREAM *stream);
/* 整数へ変換 */
PARSE_RESULT parse_int(CZSTREAM *stream, int *retval);

/* 命令の実行 */
static int search_and_call(TOKEN *token, CZSTREAM *stream);
/* 記号からオペコードへの変換 */
static enum opecode search_mark_to_opecode(TOKEN_TYPE mark);

/* 変数の解析　*/
static PARSE_RESULT parse_var(TOKEN *token, int *var_index);

#define NL_STRING_PT	0
#define TAB_STRING_PT	2

typedef void (*STATEMENTFUNC)(CZSTREAM *);
static struct __keyword_item {
	char *keyword;
	STATEMENTFUNC func;
} st_keyword_list[] = {
	{ "PRINT", parse_print },
	{ "IF", parse_if },
	{ "GOTO", parse_goto },
	{ "LET", parse_let },
	{ "INPUT", parse_input },
	{ "GOSUB", parse_gosub },
	{ "RETURN", parse_return },
	{ "CLEAR", parse_clear },
	{ "END", parse_end },
	{ NULL, NULL },
};

static struct __mark_item {
	TOKEN_TYPE mark;	/* トークンタイプ */
	enum opecode opecode;	/* オペコード */
} st_mark_list[] = {
	{ TOKEN_TYPE_PLUS, OP_ADD },		/* + */
	{ TOKEN_TYPE_MINUS, OP_SUB },		/* - */
	{ TOKEN_TYPE_MUL, OP_MUL },		/* * */
	{ TOKEN_TYPE_DIV, OP_DIV },		/* / */
	{ TOKEN_TYPE_EQ, OP_CMP_EQ },		/* = */
	{ TOKEN_TYPE_GT, OP_CMP_GT },		/* > */
	{ TOKEN_TYPE_GE, OP_CMP_GE },		/* >= */
	{ TOKEN_TYPE_LT, OP_CMP_LT },		/* < */
	{ TOKEN_TYPE_LE, OP_CMP_LE },		/* <= */
	{ TOKEN_TYPE_NE, OP_CMP_NE },		/* <> */
	{ TOKEN_TYPE_NONE, OP_NOP },
};

/* 記号からオペコードへの変換 */
static enum opecode search_mark_to_opecode(TOKEN_TYPE mark)
{
	int i;
	i = 0;
	while(st_mark_list[i].mark != TOKEN_TYPE_NONE) {
		if(st_mark_list[i].mark == mark) {
			break;
		}
		i++;
	}
	return st_mark_list[i].opecode;
}
	

/* プログラムメモリへコード追加 */
static void add_progmem(int data)
{
	g_vmmem[g_pc] = data;
	g_pc++;
	g_vmmem_last = g_pc;
	if(g_pc >= VMMEM_MAX) {
		ErrorExit(ERR_PROGMEM_OVERFLOW, st_now_line_num, st_now_basic_line_num);
	}
}

/* 文字列テーブルへの文字列追加 */
static void add_string(char *str)
{
	int i = 0;

	while(str[i] != 0) {
		g_string_pool[st_string_pt] = str[i];
		i++;
		st_string_pt++;
		if(st_string_pt >= STRING_POOL_MAX) {
			ErrorExit(ERR_STRINGMEM_OVERFLOW, st_now_line_num, st_now_basic_line_num);
		}
		g_string_pool[st_string_pt] = 0;
	}
	st_string_pt++;
}

/* 行番号テーブルへ追加 */
static void add_line_num_table(int line_num, int address)
{
	int i;

	for(i = 0; i < st_table_pt; i++) {
		if(g_line_num_table[i].line_num == line_num) {
			ErrorExit(ERR_ALREADY_DEFINED_LINE_NUM, st_now_line_num, st_now_basic_line_num);
		}
	}
	g_line_num_table[st_table_pt].line_num = line_num;
	g_line_num_table[st_table_pt].address = address;
	st_table_pt++;
	if(st_table_pt >= LINE_NUM_MAX) {
		ErrorExit(ERR_LINENUMMEM_OVERFLOW, st_now_line_num, st_now_basic_line_num);
	}
	g_line_num_table[st_table_pt].line_num = -1;
	g_line_num_table[st_table_pt].address = -1;
}

void parser(CZSTREAM *stream)
{
	g_pc = 0;
	st_string_pt = 0;
	st_table_pt = 0;
	st_now_line_num = 1;
	add_string("\n");
	add_string("\t");
	while(!stream->eof(stream)) {
		st_now_basic_line_num = -1;
		parse_line(stream);
		st_now_line_num++;
	}
	add_progmem(OP_END);
}

/* 行解析 */
static void parse_line(CZSTREAM *stream)
{
	TOKEN *token;

	token = lex_get_token(stream);
	if(token->token_type == TOKEN_TYPE_EOF) {
		return;
	}

	if(token->token[0] == '#') {	 /* コメント */
		/* 改行までスキップ */
		do {
			token = lex_get_token(stream);
			if(token == NULL) {
				return;
			}
		} while(token->token_type != TOKEN_TYPE_NL);
	} else if(isdigit((int)token->token[0])) {	/* 数値なら */
		int line_num;
		/* 行番号 */
		if(conv_unsigned_int(token, &line_num) != PARSE_RESULT_TRUE) {
			ErrorExit(ERR_LINENUM, st_now_line_num, st_now_basic_line_num);
		}
		add_line_num_table(line_num, g_pc);
		st_now_basic_line_num = line_num;

		/* 行番号を設定する命令を追加 */
		add_progmem(OP_LD_EL);
		add_progmem(st_now_line_num);
		add_progmem(OP_LD_BL);
		add_progmem(st_now_basic_line_num);
		
		/* 命令解析 */
		parse_statement(stream);
		/* CR */
		token = lex_get_token(stream);
		if(token == NULL) {
			return;
		}
		if(token->token_type != TOKEN_TYPE_NL) {
			ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
		}
	} else {
		ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
	}
}

/* tokenを符号無し整数に変換 */
static PARSE_RESULT conv_unsigned_int(TOKEN *token, int *retval)
{
	int pt = 0;

	*retval = 0;
	if(token->token[0] == '0' && token->token[1] != 0) {
		return PARSE_RESULT_FALSE;
	}
	while(token->token[pt] != 0) {
		if(!isdigit((int)token->token[pt])) {
			return PARSE_RESULT_FALSE;
		}
		*retval = *retval * 10 + token->token[pt] - '0';
		if(*retval & ~0x7fff) {
			return PARSE_RESULT_OVERFLOW;
		}
		pt++;
	}
	return PARSE_RESULT_TRUE;
}


/* 命令解析 */
static void parse_statement(CZSTREAM *stream)
{
	TOKEN *token;

	token = lex_get_token(stream);
	if(!search_and_call(token, stream)) {	/* 予約語解析して実行 */
		/* 予約語以外なら */
		lex_unget_token(token);
		/* 変数へ数値代入として解析する */
		parse_let(stream);
	}
}

/* 命令ごとの解析ルーチンを呼ぶ */
static int search_and_call(TOKEN *token, CZSTREAM *stream)
{
	int i = 0;

	while(st_keyword_list[i].keyword != NULL) {
		if(Strcmp(token->token, st_keyword_list[i].keyword) == 0) {
			st_keyword_list[i].func(stream);
			return 1;
		}
		i++;
	}

	return 0;
}

/* PRINT解析 */
static void parse_print(CZSTREAM *stream)
{
	TOKEN *token;
	int add_nl = 0;

	while(1) {
		token = lex_get_token(stream);
		if(token == NULL) {
			add_nl = 0;
			break;
		} else if(token->token_type == TOKEN_TYPE_STRING) { /* 文字列なら */
			add_progmem(OP_PRN_STRING);
			add_progmem(st_string_pt);
			add_string(token->token);
			add_nl = 1;
		} else if(token->token_type == TOKEN_TYPE_MARK) { /* 記号なら */
			if(token->token[0] == ',' && token->token[1] == 0) {
				add_progmem(OP_PRN_STRING);
				add_progmem(TAB_STRING_PT);
				add_nl = 1;
			} else if(token->token[0] == ';' && token->token[1] == 0) {
				add_nl = 0;
			} else {
				ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
			}
		} else if(token->token_type != TOKEN_TYPE_EOF && token->token_type != TOKEN_TYPE_NL) {
			lex_unget_token(token);
			parse_expression(stream);
			add_progmem(OP_PRN);
			add_nl = 1;
		}  else {
			lex_unget_token(token);
			break;
		}
	}

	if(add_nl) {
		add_progmem(OP_PRN_STRING);
		add_progmem(NL_STRING_PT);
	}
}

/* IF解析 */
static void parse_if(CZSTREAM *stream)
{
	int jmp_addr;
	TOKEN *token;

	/* 式解析 */
	parse_expression(stream);
	/* 偽ならJMPのオペコードを追加 */
	add_progmem(OP_NCOND_JMP);
	jmp_addr = g_pc;
	/* 飛び先のアドレスを仮設定 */
	add_progmem(0);
	
	token = lex_get_token(stream);
	if(Strcmp(token->token, "then") != 0) { /* THENでなければ */
		ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
	}
	/* 次の命令を解析 */
	parse_statement(stream);
	/* 負のときの飛び先を再設定 */
	g_vmmem[jmp_addr] = g_pc;
}

/* GOTO解析 */
static void parse_goto(CZSTREAM *stream)
{
	/* 式解析 */
	parse_expression(stream);
	add_progmem(OP_JMP_NUM);
}

/* LET解析 */
static void parse_let(CZSTREAM *stream)
{
	TOKEN *token = lex_get_token(stream);
	int var_index = 0;

	/* 変数 */
	if(strlen(token->token) != 1) {
		ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
	}
	if(!isalpha((int)token->token[0])) {
		ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
	}
	PARSE_RESULT res = parse_var(token, &var_index);
	if(res == PARSE_RESULT_FALSE) {
		ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
	}
	/* = */
	token = lex_get_token(stream);
	if(token->token_type != TOKEN_TYPE_EQ) {
		ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
	}
	/* 式解析 */
	parse_expression(stream);

	add_progmem(OP_POP_VAR);
	add_progmem(var_index);
}

/* INPUT解析 */
static void parse_input(CZSTREAM *stream)
{
	TOKEN *token;
	int var_index = 0;
	PARSE_RESULT res;
	int count;

	count = 0;
	while(1) {
		/* 変数の解析 */
		token = lex_get_token(stream);
		res  = parse_var(token, &var_index);
		if(res != PARSE_RESULT_TRUE) {
			ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
		}
		add_progmem(OP_PUSH_INT);
		add_progmem(var_index);
		count++;

		/* カンマの解析 */
		token = lex_get_token(stream);
		if(token->token[0] != ',' || token->token[1] != 0) {
			lex_unget_token(token);
			break;
		}
	}
	add_progmem(OP_INPUT);
	add_progmem(count);
}

/* GOSUB解析 */
static void parse_gosub(CZSTREAM *stream)
{
	/* 式解析 */
	parse_expression(stream);
	add_progmem(OP_CALL);
}

/* RETURN解析 */
static void parse_return(CZSTREAM *stream)
{
	add_progmem(OP_RET);
}

/* CLEAR解析 */
static void parse_clear(CZSTREAM *stream)
{
	add_progmem(OP_CLEAR);
}

/* END解析 */
static void parse_end(CZSTREAM *stream)
{
	add_progmem(OP_END);
}

/* 式解析 */
static void parse_expression(CZSTREAM *stream)
{
	TOKEN *token;

	/* 項1解析 */
	parse_item1(stream);

	while(1) {
		token = lex_get_token(stream);
		if(token->token_type != TOKEN_TYPE_EQ &&
			token->token_type != TOKEN_TYPE_LT && 
			token->token_type != TOKEN_TYPE_GT &&
			token->token_type != TOKEN_TYPE_LE &&
			token->token_type != TOKEN_TYPE_GE &&
			token->token_type != TOKEN_TYPE_NE) {
			lex_unget_token(token);
			break;
		}
		TOKEN_TYPE backup_type = token->token_type;
		
		/* 項1解析 */
		parse_item1(stream);

		/* オペコードを追加 */
		enum opecode ope = search_mark_to_opecode(backup_type);
		if(ope == OP_NOP) {
			ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
		}
		add_progmem(ope);
	}
}

/* 文字列比較 */
static int Strcmp(const char *str1, const char *str2)
{
	int i;

	if(strlen(str1) != strlen(str2)) {
		return -1;
	}

	for(i = 0; i < strlen(str1); i++) {
		int ch1 = toupper((int)str1[i]);
		int ch2 = toupper((int)str2[i]);
		if(ch1 != ch2) {
			return -1;
		}
	}
	return 0;
}

/* 項1解析 */
static void parse_item1(CZSTREAM *stream)
{
	/* 項2解析 */
	parse_item2(stream);
	while(1) {
		TOKEN *token = lex_get_token(stream);
		if(token->token_type != TOKEN_TYPE_PLUS &&
			token->token_type != TOKEN_TYPE_MINUS) {
			lex_unget_token(token);
			break;
		}
		TOKEN_TYPE backup_type = token->token_type;
	
		/* 項2解析 */
		parse_item2(stream);

		enum opecode ope = search_mark_to_opecode(backup_type);
		if(ope == OP_NOP) {
			ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
		}

		/* プログラムメモリへ追加 */
		add_progmem(ope);
	}
}

/* 項2解析 */
static void parse_item2(CZSTREAM *stream)
{
	/* 基本式解析 */
	parse_basis_expression(stream);
	while(1) {
		TOKEN *token = lex_get_token(stream);
		if(token->token_type != TOKEN_TYPE_MUL &&
			token->token_type != TOKEN_TYPE_DIV) {
			lex_unget_token(token);
			break;
		}
		TOKEN_TYPE backup_type = token->token_type;
	
		/* 基本式解析 */
		parse_basis_expression(stream);

		enum opecode ope = search_mark_to_opecode(backup_type);
		if(ope == OP_NOP) {
			ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
		}

		/* プログラムメモリへ追加 */
		add_progmem(ope);
	}
}

/* 基本式解析 */
static void parse_basis_expression(CZSTREAM *stream)
{
	TOKEN *token = lex_get_token(stream);

	if(token->token_type == TOKEN_TYPE_LPARENTHESES) { /* "("なら */
		/* 式解析 */
		parse_expression(stream);
		
		token = lex_get_token(stream);

		if(token->token_type != TOKEN_TYPE_RPARENTHESES) { /* ")"でないなら */
			ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
		}
	} else {
		/* 整数へ変換 */
		PARSE_RESULT ret;
		int retval;
		lex_unget_token(token);
		ret = parse_int(stream, &retval);
		if(ret == PARSE_RESULT_FALSE) { /* 整数でないなら */
			int var_index;
			/* 変数 */
			if(strlen(token->token) != 1) {
				ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
			}
			if(!isalpha((int)token->token[0])) {
				ErrorExit(ERR_SYNTAX, st_now_line_num, st_now_basic_line_num);
			}
			if(token->token[0] >= 'a') {
				var_index = token->token[0] - 'a';
			} else {
				var_index = token->token[0] - 'A';
			}
			add_progmem(OP_PUSH_VAR);
			add_progmem(var_index);
		} else if(ret == PARSE_RESULT_TRUE) {	/* 整数なら */
			add_progmem(OP_PUSH_INT);
			add_progmem(retval);
		} else {
			ErrorExit(ERR_OVERFLOW, st_now_line_num, st_now_basic_line_num);
		}
	}
}

/* 整数へ変換 */
PARSE_RESULT parse_int(CZSTREAM *stream, int *retval)
{
	TOKEN *token;

	token = lex_get_token(stream);
	if(token->token_type == TOKEN_TYPE_MINUS) { /* "-"なら */
		*retval = -1;
		token = lex_get_token(stream);
	} else if(token->token_type == TOKEN_TYPE_PLUS) { /* "+"なら */
		*retval = 1;
		token = lex_get_token(stream);
	} else {
		*retval = 1;
	}
	if(token->token_type == TOKEN_TYPE_EOF || token->token_type == TOKEN_TYPE_NL) {
		return PARSE_RESULT_FALSE;
	}
	int val;
	PARSE_RESULT res = conv_unsigned_int(token, &val);
	if(res == PARSE_RESULT_TRUE) {
		*retval = *retval * val;
	}
	
	return res;
}

/* 変数の解析　*/
static PARSE_RESULT parse_var(TOKEN *token, int *var_index)
{
	if(token->token[1] != 0) {
		return PARSE_RESULT_FALSE;
	}
	if(token->token[0] >= 'a' && token->token[1] <= 'z') {
		*var_index = token->token[0] - 'a';
		return PARSE_RESULT_TRUE;
	}
	if(token->token[0] >= 'A' && token->token[1] <= 'Z') {
		*var_index = token->token[0] - 'A';
		return PARSE_RESULT_TRUE;
	}
	return PARSE_RESULT_FALSE;
}
